import time

__all__ = ('Limiter',)


class Limiter:
    """An implementation of the leaky bucket algorithm.

    Args:
        rate (int): The max number of tokens available.
        per (float): How fast the tokens are restored in seconds.

    """
    __slots__ = ('rate', 'per', 'leak_rate', '_last', '_tokens')

    def __init__(self, rate, per):
        self.rate = int(rate)
        self.per = float(per)
        self.leak_rate = self.per / self.rate
        self._last = 0.0
        self._tokens = rate

    def get_tokens(self, current=None):
        current = time.monotonic() if current is None else current
        last = self._last or current

        tokens = self._tokens

        if current > last + self.per:
            # Passed window
            tokens = self.rate
        else:
            # Add back any tokens leaked
            elapsed = current - last
            tokens += int(elapsed // self.leak_rate)

        return tokens

    def get_retry_after(self, current=None, *, tokens=None):
        current = time.monotonic() if current is None else current
        last = self._last or current

        if tokens is None:
            tokens = self.get_tokens(current)
        if tokens:
            return 0

        elapsed = current - last
        return max(0, self.leak_rate - elapsed)

    def update_rate_limit(self, current=None):
        current = time.monotonic() if current is None else current
        if not self._last:
            self._last = current

        tokens = self.get_tokens(current)

        retry_after = 0
        if tokens:
            self._tokens = tokens - 1
        else:
            retry_after = self.get_retry_after(current, tokens=tokens)

        # Only update _last when a token is leaked
        if current - self._last > self.leak_rate:
            self._last = current

        return retry_after

    def reset(self):
        self._last = 0.0
        self._tokens = self.rate



import asyncio
import builtins
import datetime
import functools
import inspect
import json
import logging
import sys
from typing import Dict, Iterable, List, Optional, Union
from urllib.parse import quote

import aiohttp

from .datapoint import DataPoint, Resolution
from .errors import HTTPException
from .iterators import AsyncPlayerListIterator, AsyncSessionIterator
from .limiter import Limiter
from .player import IdentifierType, Player
from .server import Server
from . import __version__, utils

__all__ = ('BattleMetricsClient',)

log = logging.getLogger(__name__)


async def _json_or_text(response):
    text = await response.text(encoding='utf-8')
    if response.headers.get('content-type') == 'application/json':
        return json.loads(text)
    return text


class _MaybeUnlock:
    def __init__(self, lock: asyncio.Lock):
        self.lock = lock
        self._unlock = True

    async def __aenter__(self):
        await self.lock.acquire()
        return self

    def defer(self, unlock_after: Optional[float] = None):
        self._unlock = False
        if unlock_after is not None:
            loop = asyncio.get_running_loop()
            return loop.call_later(unlock_after, self.lock.release)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._unlock:
            self.lock.release()


class _Route:
    BASE = 'https://api.battlemetrics.com'

    def __init__(self, method, path, **params):
        self.path = path
        self.method = method
        url = self.BASE + path
        if params:
            url = url.format(**{k: quote(v) if isinstance(v, str) else v for k, v in params.items()})
        self.url = url


def _alias_param(name, alias):
    """Alias a keyword parameter in a function.
    Throws a TypeError when a value is given for both the
    original kwarg and the alias.
    """
    MISSING = object()

    def deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            alias_value = kwargs.pop(alias, MISSING)
            if alias_value is not MISSING:
                if name in kwargs:
                    raise TypeError(f'Cannot pass both {name!r} and {alias!r} in call')
                kwargs[name] = alias_value
            return func(*args, **kwargs)
        return wrapper

    return deco


def _get_my_name():
    """Get the name of the function that called this."""
    return inspect.stack()[1].function


def _add_bucket(rate, per):
    """Apply a bucket to check when making an API request."""
    def deco(func):
        added_bucket = False

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            nonlocal added_bucket
            if not added_bucket:
                self._buckets[func.__name__] = Limiter(rate, per)
                added_bucket = True

            return func(self, *args, **kwargs)

        return wrapper
    return deco


class BattleMetricsClient:
    """An interface to the battlemetrics API.

    Args:
        session (aiohttp.ClientSession): The session to make requests with.
        token (Optional[str]): An optional authorization token to use when
            making requests.
        sleep_on_ratelimit (bool): Whether ratelimits should be handled by
            sleeping or raising an error.

    """
    _Route = _Route

    def __init__(
        self,
        session: aiohttp.ClientSession,
        token: Optional[str] = None,
        *,
        sleep_on_ratelimit: bool = True
    ):
        self.session = session
        self.token = token
        self.sleep_on_ratelimit = sleep_on_ratelimit
        _user_agent = 'https://github.com/thegamecracks/abattlemetrics ({0}) Python/{1[0]}.{1[1]} aiohttp/{2}'
        self._user_agent = _user_agent.format(__version__, sys.version_info, aiohttp.__version__)
        self._reqlock = asyncio.Lock()
        self._buckets = {}

    async def _request(
            self, route: _Route, *,
            bucket: Optional[str] = None,
            params: Optional[dict] = None,
            **kwargs):
        headers = {
            'User-Agent': self._user_agent
        }

        if self.token:
            headers['Authorization'] = f'Bearer {self.token}'

        if 'json' in kwargs:
            headers['Content-Type'] = 'application/json'
            kwargs['data'] = json.dumps(
                kwargs.pop('json'), separators=(',', ':'), ensure_ascii=True)

        if params is None:
            params = {}

        bucket = self._buckets.get(bucket)

        async with _MaybeUnlock(self._reqlock) as lock:
            for tries in range(5):
                if bucket is not None:
                    retry_after = bucket.get_retry_after()
                    if retry_after:
                        log.debug(
                            '{0.method} {0.url} locally rate limited for {1:.2f}s'.format(
                                route, retry_after)
                        )
                        await asyncio.sleep(retry_after)

                async with self.session.request(
                        route.method, route.url,
                        headers=headers, params=params, **kwargs) as r:
                    if bucket is not None:
                        bucket.update_rate_limit()

                    log.debug(f'{route.method} {route.url} returned {r.status}')
                    data = await _json_or_text(r)

                    retry_after = r.headers.get('Retry-After', 0)
                    retry_after = float(retry_after)
                    # NOTE: battlemetrics doesn't always give
                    # X-Rate-Limit-Remaining header
                    if retry_after:
                        if self.sleep_on_ratelimit:
                            log.warning(f'Rate limited; retrying in {retry_after:.2f}')
                            await asyncio.sleep(retry_after)
                            log.debug('Done sleeping for rate limit, retrying...')
                            continue
                        else:
                            e = HTTPException(r, data)
                            # unlock once rate limit is done
                            log.warning(
                                'Rate limited; sleep_on_ratelimit '
                                'is False, raising exception', exc_info=e
                            )
                            lock.defer(retry_after)
                            raise e

                    if r.status != 200:
                        e = HTTPException(r, data)
                        log.exception(
                            'Response %d caused with:\nRoute: %s %s\nParams: %s',
                            r.status, route.method, route.path, params, exc_info=e
                        )
                        raise e

                    return data

            # No more retries left
            raise HTTPException(r, data)

    @_alias_param('stop', 'before')
    @_alias_param('start', 'after')
    async def get_player_count_history(
        self,
        server_id: int,
        *,
        start: datetime.datetime, stop: datetime.datetime,
        resolution: Optional[Resolution] = Resolution.RAW
    ) -> List[DataPoint]:
        """Obtain a server's player count history.

        Args:
            server_id (int): The server's ID.
            start (datetime.datetime):
                Get the player count history after this time.
                Naive datetimes are assumed to be in local time.
                This parameter has "after" as an alias.
            stop (datetime.datetime):
                Get the player count history before this time.
                Naive datetimes are assumed to be in local time.
                This parameter has "before" as an alias.
            resolution (Optional[Resolution]):
                The resolution of the data points. If raw, the data points
                will only have value provided. Any other option will provide
                value, min, and max.

        Returns:
            List[DataPoint]: A list of data points sorted by timestamp.

        """
        r = _Route('GET', '/servers/{server_id}/player-count-history',
                   server_id=int(server_id))

        params = {
            'start': utils.isoify_datetime(start),
            'stop': utils.isoify_datetime(stop),
            'resolution': resolution.value
        }

        payload = await self._request(r, params=params)
        data = payload['data']

        datapoints = [DataPoint(d['attributes']) for d in data]
        datapoints.sort(key=lambda dp: dp.timestamp)

        return datapoints

    def get_player_session_history(
        self,
        player_id: int,
        *,
        limit: int,
        organization_ids: Optional[Iterable[int]] = None,
        server_ids: Optional[Iterable[int]] = None,
        include_servers: bool = True,
    ) -> AsyncSessionIterator:
        """Return an async iterator yielding a player's sessions
        ordered by most recent.

        Args:
            limit (int): The maximum number of sessions to yield.
            organization_ids (Optional[Iterable[int]]): A list of
                organization IDs to filter sessions by.
            server_ids (Optional[Iterable[int]]): A list of
                server IDs to filter sessions by.
            include_servers (bool): Request server data as well,
                filling in the `server` attribute of each Session object.

        Returns:
            AsyncSessionIterator

        """
        limit = int(limit)
        player_id = int(player_id)

        if limit < 1:
            raise ValueError('limit must be at least 1')

        params = {}
        if organization_ids:
            params['filter[organizations]'] = ','.join([
                str(int(n)) for n in organization_ids
            ])
        if server_ids:
            params['filter[servers]'] = ','.join([
                str(int(n)) for n in server_ids
            ])

        include = []
        if include_servers:
            include.append('server')
        include = ','.join(include)
        if include:
            params['include'] = include

        return AsyncSessionIterator(self, limit, player_id, params)

    @_alias_param('stop', 'before')
    @_alias_param('start', 'after')
    async def get_player_time_played_history(
        self,
        player_id: int,
        server_id: int,
        *,
        start: datetime.datetime,
        stop: datetime.datetime
    ) -> List[DataPoint]:
        """Obtain a player's time played history for a server.

        Start and stop are truncated to the date.

        Args:
            player_id (int): The player's ID.
            server_id (int): The server ID.
            start (datetime.datetime):
                Get the time played history after this time.
                Naive datetimes are assumed to be in local time.
                This parameter has "after" as an alias.
            stop (datetime.datetime):
                Get the time played history before this time.
                Naive datetimes are assumed to be in local time.
                This parameter has "before" as an alias.

        Returns:
            List[DataPoint]: A list of data points sorted by timestamp.
                Each data point is per day and their values are in seconds.

        """
        r = _Route('GET', '/players/{player_id}/time-played-history/{server_id}',
                   player_id=int(player_id), server_id=int(server_id))

        params = {
            'start': utils.isoify_datetime(start),
            'stop': utils.isoify_datetime(stop)
        }

        payload = await self._request(r, params=params)
        data = payload['data']

        datapoints = [DataPoint(d['attributes']) for d in data]
        datapoints.sort(key=lambda dp: dp.timestamp)

        return datapoints

    def list_players(
        self,
        *,
        limit: int = 10,
        countries: Optional[Iterable[str]] = None,
        distance: Optional[int] = None,
        first_seen_after: Optional[datetime.datetime] = None,
        first_seen_before: Optional[datetime.datetime] = None,
        game: Optional[str] = None,
        include_identifiers: bool = False,
        is_online: Optional[bool] = None,
        last_seen_after: Optional[datetime.datetime] = None,
        last_seen_before: Optional[datetime.datetime] = None,
        online_at: Optional[datetime.datetime] = None,
        organization_id: Optional[int] = None,
        public: bool = True,
        search: Optional[str] = None,
        server_ids: Optional[Iterable[int]] = None
    ) -> AsyncPlayerListIterator:
        """Search records for players ordered by most recent.

        Player objects returned will not have metadata such as `first_time`.

        Certain parameters are documented with
        their authentication requirements.

        Args:
            limit (int): The maximum number of players to yield.
            countries (Optional[Iterable[str]]):
                Filter by a list of ISO 3166-1 alpha-2 country codes.
            distance (Optional[int]):
                Filter by maximum server distance to the client in kilometres.
            first_seen_after (Optional[datetime.datetime]):
                Filter by players first seen after this datetime.
                Naive datetimes are assumed to be in local time.
                `server_ids` must also be provided for this parameter.
                Requires token with "View RCON information" permission.
            first_seen_before (Optional[datetime.datetime]):
                Filter by players first seen before this datetime.
                Naive datetimes are assumed to be in local time.
                `server_ids` must also be provided for this parameter.
                Requires token with "View RCON information" permission.
            game (Optional[str]):
                Filter by this game name.
            include_identifiers (bool): If True, fetches and fills in
                the `identifiers` atttribute for each Player.
                For public results, this will only contain identifers
                of type `IdentifierType.NAME`.
            is_online (Optional[bool]):
                If True, only return players that are currently online.
            last_seen_after (Optional[datetime.datetime]):
                Filter by players last seen after this datetime.
                Naive datetimes are assumed to be in local time.
            last_seen_before (Optional[datetime.datetime]):
                Filter by players last seen before this datetime.
                Naive datetimes are assumed to be in local time.
            online_at (Optional[datetime.datetime]):
                Filter by players being online at this datetime.
                Naive datetimes are assumed to be in local time.
                `public` must also be False.
            organization_id (Optional[int]):
                Filter by an organization ID.
                Requires authentication.
            public (bool): If True, includes public records.
                Otherwise, only players in your servers are returned.
                Requires authentication for non-public results.
            search (Optional[str]): The search term to match each player to.
                Explanation on this is provided here:
                https://www.battlemetrics.com/players
                Note that public results only allow you to match names.
            server_ids (Optional[Iterable[int]]):
                Filter by a list of server IDs.

        Returns:
            AsyncPlayerListIterator

        """
        limit = int(limit)
        if limit < 1:
            raise ValueError('limit must be at least 1')

        params = {'sort': '-lastSeen'}
        if countries:
            params['filter[server][countries][]'] = [str(c) for c in countries]
        if distance:
            distance = int(distance)
            if distance < 0:
                raise ValueError('distance cannot be negative')
            params['filter[server][maxDistance]'] = distance
        if first_seen_after or first_seen_before:
            p = 'after' if first_seen_after else 'before'
            if not self.token:
                raise ValueError(f'authentication required for first_seen_{p}')
            elif not server_ids:
                # Will result in 500
                raise ValueError(f'server_ids required for first_seen_{p}')

            params['filter[firstSeen]'] = '{}:{}'.format(
                utils.isoify_datetime(first_seen_after) if first_seen_after else '',
                utils.isoify_datetime(first_seen_before) if first_seen_before else ''
            )
        if game:
            params['filter[server][game]'] = str(game)

        include = []
        if include_identifiers:
            include.append('identifier')
        if include:
            params['include'] = ','.join(include)
        if is_online:
            params['filter[online]'] = 'true'
        if last_seen_after:
            params['filter[after]'] = utils.isoify_datetime(last_seen_after)
        if last_seen_before:
            params['filter[before]'] = utils.isoify_datetime(last_seen_before)
        if online_at:
            if public:
                raise ValueError('public=False required for online_at')
            params['filter[sessions][at]'] = utils.isoify_datetime(online_at)
        if organization_id:
            if not self.token:
                raise ValueError('authentication required for organization_id')
            params['filter[organization]'] = int(organization_id)
        if not public:
            if not self.token:
                raise ValueError('authentication required for public=False')
            params['filter[public]'] = 'false'  # API defaults to true
        if search:
            params['filter[search]'] = str(search)
        if server_ids:
            params['filter[servers]'] = ','.join([
                str(int(n)) for n in server_ids
            ])

        return AsyncPlayerListIterator(self, limit, params)

    @_add_bucket(1, 1)
    async def match_players(
        self,
        *identifiers: Union[int, str],
        type: IdentifierType
    ) -> Dict[Union[int, str], Optional[int]]:
        """Get the player IDs associated with the given identifiers.

        Requires authentication token with "View RCON information" permission.

        This endpoint is rate limited to 1/1s.

        Args:
            identifiers (Union[int, str]): The player identifiers.
                You are allowed to request up to 100 identifiers at a time.
            type (IdentifierType): The type of identifier being provided.

        Returns:
            Dict[Union[int, str], Optional[int]]:
                Mapping of player identifier to battlemetrics ID.
                The identifier will be the same type as passed
                in the parameters.
                Unmatched identifiers are mapped to None.

        """
        def transfer_type(x):
            return identifier_types[x](x)

        if not identifiers:
            raise TypeError('At least 1 identifier must be given')
        elif len(identifiers) > 100:
            raise ValueError('Only 100 identifiers can be requested at once')

        r = _Route('POST', '/players/match')
        data = {
            'data': [
                {
                    'type': 'identifier',
                    'attributes': {
                        'type': type.value,
                        'identifier': str(i)
                    }
                }
                for i in identifiers
            ]
        }
        payload = await self._request(r, json=data, bucket=_get_my_name())
        data = payload['data']
        if not data:
            return {}

        identifier_types = {str(i): builtins.type(i) for i in identifiers}

        results = dict.fromkeys(identifiers)
        for d in data:
            i = transfer_type(d['attributes']['identifier'])
            results[i] = int(d['relationships']['player']['data']['id'])
        return results

    async def get_player_info(self, player_id: int) -> Player:
        """Get a player's info from their battlemetrics ID.

        Args:
            player_id (int): The player's ID.

        Returns:
            Player

        """
        r = _Route('GET', '/players/{player_id}', player_id=int(player_id))
        payload = await self._request(r)
        return Player(payload['data'])

    async def get_server_info(
        self,
        server_id: int,
        *,
        include_players=False
    ) -> Server:
        """Obtain server info given an ID.

        Args:
            server_id (int): The server's ID.
            include_players (bool): Whether to also fetch player data.
                This affects the `players` attribute.

        Returns:
            Server

        """
        r = _Route('GET', '/servers/{server_id}', server_id=int(server_id))

        params = {}

        include = []
        if include_players:
            include.append('player')

        include = ','.join(include)
        if include:
            params['include'] = include

        return Server(await self._request(r, params=params))



from dataclasses import dataclass, field
import datetime
import functools
import types
from typing import Optional

from .mixins import PayloadIniter
from .server import Server
from . import utils

__all__ = ('Session',)


def _parse_optional_datetime(date_string: Optional[str]):
    if not date_string:
        return None
    return utils.parse_datetime(date_string)


@dataclass(frozen=True, init=False)
class Session(PayloadIniter):
    """Represents a player session.

    Attributes:
    first_time (bool):
        Whether this is the first time the player is on the server.
    id (str): The session identifier.
    player_id (int): The player's ID.
    player_name (str): The player's name.
    payload (dict): A read-only view of the raw payload.
    playtime (float): How long this session has lasted in seconds.
        If `start` or `stop` is None, then this will be 0.
    server (Optional[Server]): The server this session took place in.
        This may be None if the server data was not included.
    server_id (int): The ID of the server this session took place in.
    start (Optional[datetime.datetime]):
        The start of the session as an aware datetime.
        Battlemetrics may occasionally not provide this date.
    stop (Optional[datetime.datetime]):
        The end of the session as an aware datetime.
        Battlemetrics may occasionally not provide this date.

    """
    __init_attrs = (
        {'name': 'first_time', 'path': ('attributes', 'firstTime')},
        'id',
        {'name': 'player_id', 'type': int,
         'path': ('relationships', 'player', 'data', 'id')},
        {'name': 'player_name', 'path': ('attributes', 'name')},
        {'name': 'server_id', 'type': int,
         'path': ('relationships', 'server', 'data', 'id')},
        {'name': 'start', 'type': _parse_optional_datetime,
         'path': ('attributes', 'start')},
        {'name': 'stop', 'type': _parse_optional_datetime,
         'path': ('attributes', 'stop')}

    )

    first_time: bool                   = field(hash=False, repr=False)
    id: str
    payload: dict                      = field(hash=False, repr=False)
    player_id: int                     = field(hash=False)
    player_name: str                   = field(hash=False, repr=False)
    server: Optional[Server]           = field(hash=False, repr=False)
    server_id: int                     = field(hash=False)
    start: Optional[datetime.datetime] = field(hash=False, repr=False)
    stop: Optional[datetime.datetime]  = field(hash=False, repr=False)

    def __init__(self, payload):
        super().__setattr__('payload', types.MappingProxyType(payload))
        # NOTE: server is manually assigned by AsyncSessionIterator

        self.__init_attrs__(payload, self.__init_attrs)

    @functools.cached_property
    def playtime(self) -> float:
        if self.start is None or self.stop is None:
            return 0
        return (self.stop - self.start).total_seconds()



import collections
import logging
from typing import List, Optional, Union
import urllib.parse as urlparse

from .player import Player
from .session import Session
from .server import Server

__all__ = ('AsyncIterator', 'AsyncPlayerListIterator', 'AsyncSessionIterator')

log = logging.getLogger(__name__)


def _convert_params_to_array(params: dict) -> Union[dict, list]:
    """Convert a dictionary to a list of key-value pairs, unpacking
    list values to their own keys. If none of the values are a list,
    returns the dictionary untouched.
    """
    if not any(isinstance(v, list) for v in params.values()):
        return params

    new = []
    for k, v in params.items():
        if isinstance(v, list):
            for v_v in v:
                new.append((k, v_v))
        else:
            new.append((k, v))
    return new


class AsyncIterator:
    async def flatten(self):
        x = []
        async for item in self:
            x.append(item)
        return x

    async def next(self):
        raise NotImplementedError

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await self.next()


class AsyncPaginatedIterator(AsyncIterator):
    _MAX_SIZE = 100

    def __init__(self, client, route, limit, params):
        self._client = client
        self._route = route
        self._limit = limit
        self._params = params
        self._current_page = None

    def _get_next_params(self, r: dict) -> Optional[dict]:
        params = r['links'].get('next')
        if params:
            return urlparse.parse_qs(urlparse.urlparse(params).query)

    async def _request_page(self, params: Union[dict, list]):
        params = _convert_params_to_array(params)
        res = await self._client._request(self._route, params=params)
        page = await self._parse_response(res)

        params = None
        if page:
            params = self._get_next_params(res)

        self._limit -= len(page)
        if params and self._limit <= 0:
            params = None

        page.reverse()
        self._current_page = page
        self._params = params

    async def _parse_response(self, r: dict) -> List[object]:
        raise NotImplementedError

    async def next(self):
        if self._current_page:
            return self._current_page.pop()
        elif self._params is not None:
            self._params['page[size]'] = min(self._limit, self._MAX_SIZE)
            await self._request_page(self._params)

        if self._current_page:
            return self._current_page.pop()
        else:
            raise StopAsyncIteration


class AsyncPlayerListIterator(AsyncPaginatedIterator):
    def __init__(self, client, limit, params):
        route = client._Route('GET', '/players')
        super().__init__(client, route, limit, params)

    async def _request_page(self, params):
        log.debug('Requesting %d players listed', params['page[size]'])
        await super()._request_page(params)

    async def _parse_response(self, r) -> List[Player]:
        identifiers = collections.defaultdict(list)
        for payload in r['included']:
            if payload['type'] == 'identifier':
                p_id = int(payload['relationships']['player']['data']['id'])
                identifiers[p_id].append(payload)

        return [
            Player(
                payload,
                identifiers.get(int(payload['id']), None)
            )
            for payload in r['data']
        ]


class AsyncSessionIterator(AsyncPaginatedIterator):
    def __init__(self, client, limit, player_id, params):
        route = client._Route(
            'GET', '/players/{player_id}/relationships/sessions',
            player_id=player_id
        )
        super().__init__(client, route, limit, params)

    async def _request_page(self, params):
        log.debug('Requesting %d sessions', params['page[size]'])
        await super()._request_page(params)

    async def _parse_response(self, r) -> List[Session]:
        servers = {
            payload['attributes']['id']: Server(payload)
            for payload in r['included']
            if payload['type'] == 'server'
        }

        page = []
        for payload in r['data']:
            session = Session(payload)

            server = servers.get(session.server_id)
            if server is not None:
                super(Session, session).__setattr__('server', server)

            page.append(session)

        return page



import aiohttp


class BattleMetricsException(Exception):
    """The base class for all battlemetrics exceptions."""


class HTTPException(BattleMetricsException):
    """Thrown when an HTTP request fails.

    Attributes:
        status (int): The HTTP status code.

    """
    def __init__(self, response: aiohttp.ClientResponse, data, *args):
        self.status = response.status

        detail = None
        if isinstance(data, dict):
            errors = data.get('errors')
            if errors:
                detail = errors[0]['detail']
        else:
            detail = data
        self.detail = detail

        super().__init__(
            '{0.status} {0.reason}'.format(response),
            *args
        )



from dataclasses import dataclass, field
import datetime
import enum
import types
from typing import Optional, Tuple

from .mixins import PayloadIniter
from . import utils

__all__ = ('IdentifierType', 'Identifier', 'Player')


class IdentifierType(enum.Enum):
    """A player identifier type."""
    BE_GUID                  = 'BEGUID'
    BE_LEGACY_GUID           = 'legacyBEGUID'
    CONAN_CHAR_NAME          = 'conanCharName'
    EGS_ID                   = 'egsID'
    FUNCOM_ID                = 'funcomID'
    IP                       = 'ip'
    MC_UUID                  = 'mcUUID'
    NAME                     = 'name'
    PLAY_FAB_ID              = 'playFabID'
    STEAM_FAMILY_SHARE_OWNER = 'steamFamilyShareOwner'
    STEAM_ID                 = 'steamID'
    SURVIVOR_NAME            = 'survivorName'

    def __repr__(self):
        return '<{0.__class__.__name__}.{0.name}>'.format(self)


@dataclass(frozen=True, init=False)
class Identifier(PayloadIniter):
    """A player identifier.

    Attributes:
    id (int): The identifier ID.
        Note that this is not the actual player identifier.
    last_seen (datetime.datetime):
        The time this identifier was last seen as an aware datetime.
    metadata (dict): A read-only view of the payload metadata.
        This is supplied for certain identifiers, e.g. IP.
    name (Optional[str]): The player identifier.
        This may be None depending on the type, e.g. IP.
    payload (dict): A read-only view of the raw payload.
    player_id (int): The player ID this identifier is associated to.
    private (bool): Indicates if this should be considered private.
    type (IdentifierType): The type of identifier this is.

    """
    __init_attrs = (
        {'name': 'id', 'type': int},
        {'name': 'last_seen', 'type': utils.parse_datetime,
         'path': ('attributes', 'lastSeen')},
        {'name': 'name', 'path': ('attributes', 'identifier'),
         'default': None},
        {'name': 'player_id', 'type': int,
         'path': ('relationships', 'player', 'data', 'id')},
        {'name': 'private', 'path': ('attributes', 'private')},
        {'name': 'type', 'path': ('attributes', 'type'), 'type': IdentifierType}
    )

    id: int
    last_seen: datetime.datetime  = field(hash=False, repr=False)
    metadata: dict                = field(hash=False, repr=False)
    name: str                     = field(hash=False)
    payload: dict                 = field(hash=False, repr=False)
    player_id: int                = field(hash=False, repr=False)
    private: bool                 = field(hash=False, repr=False)
    type: IdentifierType          = field(hash=False, repr=False)

    def __init__(self, payload):
        super().__setattr__('payload', types.MappingProxyType(payload))

        self.__init_attrs__(payload, self.__init_attrs)

        metadata = payload['attributes']['metadata'] or {}
        super().__setattr__('metadata', types.MappingProxyType(metadata))


@dataclass(frozen=True, init=False)
class Player(PayloadIniter):
    """Represents a player.

    Attributes:
    created_at (datetime.datetime): When the player was first created
        on battlemetrics as an aware datetime.
    first_time (Optional[bool]):
        Whether this is the first time the player is on the server.
    id (int): The player's id.
    identifiers (Tuple[Identifier, ...]):
        A tuple of Identifiers for this Player. Currently, this attribute
        is only provided by the `BattleMetricsClient.list_players()` endpoint.
    name (str): The player's name.
    score (Optional[int]): The player's in-game score.
    payload (dict): A read-only view of the raw payload.
    playtime (Optional[float]): How long the player has been in the server
        for their current session in seconds.
    positive_match (bool):
        When retrieved from a search with unique identifiers,
        this will be True if one of those identifiers exactly
        matches this player.
    private (bool): Indicates if the profile is private.
        Private profiles are excluded from search and player lists.
    updated_at (datetime.datetime):
        When this player was last updated as an aware datetime.

    """
    __init_attrs = (
        {'name': 'created_at', 'type': utils.parse_datetime,
         'path': ('attributes', 'createdAt')},
        {'name': 'id', 'type': int},
        {'name': 'name', 'path': ('attributes', 'name')},
        {'name': 'private', 'path': ('attributes', 'private')},
        {'name': 'positive_match', 'path': ('attributes', 'positiveMatch')},
        {'name': 'updated_at', 'type': utils.parse_datetime,
         'path': ('attributes', 'updatedAt')}
    )
    _init_meta = ({'name': 'first_time', 'path': 'firstTime'},
                  'score', {'name': 'playtime', 'path': 'time'})

    created_at: datetime.datetime       = field(hash=False, repr=False)
    first_time: Optional[bool]          = field(hash=False, repr=False)
    id: int
    identifiers: Tuple[Identifier, ...] = field(hash=False, repr=False)
    name: str                           = field(hash=False)
    score: Optional[int]                = field(hash=False, repr=False)
    payload: dict                       = field(hash=False, repr=False)
    playtime: Optional[float]           = field(hash=False, repr=False)
    positive_match: bool                = field(hash=False, repr=False)
    private: bool                       = field(hash=False, repr=False)
    updated_at: datetime.datetime       = field(hash=False, repr=False)

    def __init__(self, payload, identifiers=None):
        def flatten_meta(metadata):
            return {x['key']: x['value'] for x in metadata}

        super().__setattr__('payload', types.MappingProxyType(payload))

        self.__init_attrs__(payload, self.__init_attrs)
        metadata = flatten_meta(payload.get('meta', {}).get('metadata', {}))
        self.__init_attrs__(metadata, self._init_meta, required=False)

        if identifiers:
            identifiers = tuple(Identifier(p) for p in identifiers)
        else:
            identifiers = ()
        super().__setattr__('identifiers', identifiers)



from dataclasses import dataclass, field
import datetime
import types
from typing import Optional, Tuple

from .mixins import PayloadIniter
from .player import Player
from . import utils

__all__ = ('Server',)


@dataclass(frozen=True, init=False)
class Server(PayloadIniter):
    """Represents a server returned by get_server_info().

    Attributes:
    address (Optional[str]): The server address, e.g. play.example.com
    country (str): The country code that the server is hosted in.
    created_at (datetime.datetime):
        When the server was created on battlemetrics as an aware datetime.
    details (dict): A dict with more specific details on the server's settings,
        such as difficulty, map, or game version.
    id (int): The server's ID.
    ip (str): The IPv4 address of the server.
    max_players (int): The maximum number of players the server allows.
    name (str): The server name.
    payload (dict): A read-only view of the raw payload.
    player_count (int): The number of players in the server.
    players (tuple): A list of Player objects. This may be empty if the query
        did not specify to include player data.
    port (int): The server's port.
    private (bool): Whether the server is private or not.
    query_port (int): The server's query port.
    rank (int): The server's rank on battlemetrics's leaderboards.
    status (str): The status of the server, i.e. "online" or "offline"
    updated_at (datetime.datetime):
        When the server was last updated on battlemetrics as an aware datetime.

    """
    __init_attrs = (
        {'name': 'created_at', 'type': utils.parse_datetime,
         'path': 'createdAt'},
        'address', 'country', 'details', {'name': 'id', 'type': int}, 'ip',
        {'name': 'max_players', 'path': 'maxPlayers'}, 'name',
        {'name': 'player_count', 'path': 'players'}, 'port',
        {'name': 'query_port', 'path': 'portQuery'}, 'private',
        'rank', 'status',
        {'name': 'updated_at', 'type': utils.parse_datetime,
         'path': 'updatedAt'}
    )

    address: Optional[str]        = field(hash=False, repr=False)
    country: str                  = field(hash=False, repr=False)
    created_at: datetime.datetime = field(hash=False, repr=False)
    details: dict                 = field(hash=False, repr=False)
    id: int
    ip: str                       = field(hash=False, repr=False)
    max_players: int              = field(hash=False, repr=False)
    name: str                     = field(hash=False)
    payload: dict                 = field(hash=False, repr=False)
    player_count: int             = field(hash=False, repr=False)
    players: Tuple[Player]        = field(hash=False, repr=False)
    port: int                     = field(hash=False, repr=False)
    private: bool                 = field(hash=False, repr=False)
    query_port: int               = field(hash=False, repr=False)
    rank: int                     = field(hash=False, repr=False)
    status: str                   = field(hash=False, repr=False)
    updated_at: datetime.datetime = field(hash=False, repr=False)

    def __init__(self, payload):
        super().__setattr__('payload', types.MappingProxyType(payload))

        data = payload.get('data')
        if data:
            attrs = data['attributes']
        else:
            attrs = payload['attributes']

        self.__init_attrs__(attrs, self.__init_attrs)

        included = payload.get('included')
        players = ()
        if included:
            players = tuple(
                Player(item) for item in included
                if item['type'] == 'player'
            )
        super().__setattr__('players', players)



from typing import Tuple, Union


class PayloadIniter:
    """Provides an __init_attrs__ method for extracting attributes
    from a payload. Use __init_attrs to specify the attributes.

    """

    def __init_attrs__(
            self, attrs, mapping: Tuple[Union[str, dict], ...], *,
            required=True):
        """Extract attributes into instance variables according to mapping.

        Each attr specified in mapping can be either a string, or a dictionary.
        A single string specifies both the instance attribute's name
        and the key in the attrs dict. A dictionary specifies the attribute's
        name as `name`, an optional key or sequence of keys for navigating
        the dictionary as `path`, and an optional `type` for converting
        the value.
        Examples:
            'address'
                Reads the "address" key and assigns it to `address`
            {'name': 'address'}
                Equivalent to above
            {'name': 'max_players', 'path': 'maxPlayers'}
                Reads the "maxPlayers" key and assigns it to `max_players`
            {'name': 'steam_id', 'path': ('details', 'serverSteamId'),
             'type': int}
                Reads the "serverSteamId" key inside "details", converts it
                to an integer, and assigns it to `steam_id`
            {'name': 'score', 'default': 0}
                Reads the "score" key and if it exists, assigns it
                to `score`, otherwise the `default` value is used.
                Does not convert to `type`.
            {'name': 'identifier', 'default_factory': dict}
                Reads the "identifier" key and if it exists, assigns it
                to `identifier`, otherwise `default_factory` is called
                for a value. Does not convert to `type`.

        Args:
            attrs (dict): A dictionary of attributes to extract values from.
                This is usually a payload.
            mapping (Tuple[Dict[str, object]]):
                The mapping to follow when extracting attributes from attrs.
            required (bool): If True, all specified attributes that do not
                have a default or default_factory must exist.
                Otherwise, some keys can be missing from attrs, in
                which case their corresponding attribute is set to None.

        Raises:
            KeyError: An attribute specified in mapping was missing from attrs.
        """
        def lookup(p):
            v = attrs
            for k in p:
                try:
                    v = v[k]
                except (KeyError, TypeError):
                    if required:
                        raise KeyError(f'attrs is missing {k!r} from {v!r}')
                    return
            return v

        missing = object()

        for x in mapping:
            if isinstance(x, str):
                super().__setattr__(x, lookup((x,)))
            else:
                name, path = x['name'], x.get('path')
                if path is None:
                    path = name
                if isinstance(path, str):
                    path = (path,)

                try:
                    val = lookup(path)
                except KeyError as e:
                    val = x.get('default', missing)
                    if val is missing:
                        val = x.get('default_factory', missing)
                        if val is missing:
                            raise e
                        val = val()

                conv = x.get('type')
                if conv is not None:
                    val = conv(val)

                super().__setattr__(x['name'], val)



import datetime
import enum
from dataclasses import dataclass, field
from typing import Optional

from .mixins import PayloadIniter
from . import utils

__all__ = ('DataPoint', 'Resolution')


class Resolution(enum.Enum):
    """The resolution to use when querying DataPoints."""
    RAW = 'raw'
    SEVEN_DAYS = 30
    THIRTY_DAYS = 60
    SIX_MONTHS = 1440


@dataclass(frozen=True, init=False, repr=False)
class DataPoint(PayloadIniter):
    """A data point of a time series.

    group (Optional[int]): If multiple groups of metrics were requested,
        this will be the index of the corresponding group.
    max (Optional[int]): The minimum value of the datapoint if relevant.
        This is usually provided when resolution is not raw.
    min (Optional[int]): The minimum value of the datapoint if relevant.
        This is usually provided when resolution is not raw.
    name (Optional[str]): The name of the metric.
        Only provided when more than one metric is requested.
    timestamp (datetime.datetime):
        The data point's timestamp as an aware datetime.
    value (int): The value of the data point.

    """
    __init_attrs = ('group', 'min', 'max', 'name', 'value')

    group: Optional[int]
    min: Optional[int]
    max: Optional[int]
    name: Optional[str]
    timestamp: datetime.datetime
    value: int

    def __init__(self, payload):
        self.__init_attrs__(payload, self.__init_attrs, required=False)
        super().__setattr__('timestamp', utils.parse_datetime(payload['timestamp']))
        assert self.value is not None, 'payload is missing "value"'

    def __repr__(self):
        fields = self.__dict__
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join([
                f'{name}={value!r}' for name, value in fields.items()
                if value is not None and not name.startswith('_')
            ])
        )



import datetime

import dateutil.parser


def isoify_datetime(dt: datetime.datetime) -> str:
    """Turn a datetime into a ISO formatted string in UTC suitable
    for parameters."""
    dt = dt.astimezone(datetime.timezone.utc)
    return dt.replace(microsecond=0, tzinfo=None).isoformat() + 'Z'


def parse_datetime(date_string: str) -> datetime.datetime:
    """Parse a datetime given by battlemetrics."""
    dt = dateutil.parser.isoparse(date_string)
    if dt.tzinfo:
        return dt.astimezone(datetime.timezone.utc)
    return dt.replace(tzinfo=datetime.timezone.utc)
